package net.pricefx.logic.core;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import net.pricefx.common.api.FormulaNature;
import net.pricefx.logic.core.metadata.MetadataCache;
import net.pricefx.logic.core.model.AbstractFormula;
import net.pricefx.logic.core.model.AbstractFormulaElement;
import net.pricefx.logic.core.model.ContextParameter;
import net.pricefx.logic.core.model.ElementCombinationType;
import net.pricefx.pckg.BusinessKey;
import net.pricefx.pckg.filesystem.FS_SupplierFactory;
import net.pricefx.pckg.filesystem.FileSystemOpsBase;
import net.pricefx.pckg.memory.ReplyingSupplierFactory;
import net.pricefx.pckg.processing.ProcessingContext;
import net.pricefx.pckg.processing.ProcessingContextImpl;
import net.pricefx.pckg.processing.ProcessingException;
import net.pricefx.pckg.processing.SupplierFactory;
import net.pricefx.pckg.processing.element.UnfoldField;
import net.pricefx.pckg.processing.filter.ItemFilter;
import net.pricefx.pckg.transform.TransformCalculationLogic;
import net.pricefx.pckg.transform.TransformWorkflowFormula;
import net.pricefx.pckg.transform.Transformations;
import net.pricefx.pckg.transform.descriptor.TypeDescriptor;
import net.pricefx.pckg.transform.descriptor.TypeDescriptors;
import net.pricefx.restapiclient.PfxClient;
import net.pricefx.restapiclient.PfxClientException;
import net.pricefx.restapiclient.api.DatamartApi;
import net.pricefx.restapiclient.api.GeneralApi;
import net.pricefx.restapiclient.api.ManualPriceListApi;
import net.pricefx.restapiclient.api.PriceGridApi;
import net.pricefx.restapiclient.api.PriceListApi;
import net.pricefx.restapiclient.criteria.And;
import net.pricefx.restapiclient.criteria.CriteriaList;
import net.pricefx.restapiclient.criteria.Or;
import net.pricefx.restapiclient.model.BaseMap;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.LocalDate;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import static net.pricefx.common.api.PriceFxInterface.*;
import static net.pricefx.logic.StudioConstants.*;
import static net.pricefx.logic.core.TypeCode.*;
import static net.pricefx.restapiclient.criteria.FieldCriteria.equal;

public abstract class AbstractLogicSerializer<T> extends FileSystemOpsBase {
    private static final Logger log = LoggerFactory.getLogger(AbstractLogicSerializer.class);

    /**
     * Filename of "logic.json" file containing logic and its elements metadata.
     */
    public static final String LOGIC_JSON_FILENAME = TransformCalculationLogic.FILENAME_LOGIC;

    /**
     * Folder name to store element files.
     */
    public static final String ELEMENTS_SUBDIR = "elements";

    /**
     * Folder name containing "params.json" files.
     */
    public static final String PARAMETERS_SUBDIR = "parameters";

    /**
     * Folder name for saving logc test runs.
     */
    public static final String RESULTS_SUBDIR = "results";

    /**
     * Default suffix added to the param file names.
     */
    public static final String PARAM_FILE_SUFFIX = ".json";

    /**
     * The calculation logic unit tests folder name
     */
    public static final String LOGIC_TESTS_DIR = TransformCalculationLogic.DIRNAME + "Test";

    public static final String PARAM_STUDIO_TYPE_CODE = "PFX_STUDIO_TYPECODE";
    public static final String PARAM_STUDIO_TYPE_CODE_LABEL = "Context";
    public static final String PARAM_STUDIO_OX_NAME = "PFX_STUDIO_OX_NAME";
    public static final String PARAM_STUDIO_DS_NAME = "PFX_STUDIO_DS_NAME";
    public static final String PARAM_STUDIO_DS_LABEL = "Datasource name";
    public static final String PARAM_STUDIO_KEY = "PFX_STUDIO_KEY_";
    public static final String PARAM_PRICE_GRID_ID = "PriceGridId";
    public static final String PARAM_PRICE_LIST_ID = "PriceListId";
    public static final String PARAM_MANUAL_PRICE_LIST_ID = "ManualPriceListId";
    public static final String PARAM_SECONDARY_KEY = "SecondaryKey";
    public static final String PARAM_DATA_CHANGE_REQUEST_NAME = "DataChangeRequestName";
    public static final String PARAM_QUOTE_NAME = "QuoteName";
    public static final String PARAM_REBATE_AGREEMENT_NAME = "RebateAgreementName";
    public static final String PARAM_REBATE_TYPE_NAME = "RebateTypeName";
    public static final String PARAM_REBATE_RECORD_NAME = "RebateRecordName";
    public static final String PARAM_CONTRACT_NAME = "ContractName";
    public static final String PARAM_CONTRACT_TERM_TYPE_NAME = "ContractTermTypeName";
    public static final String PARAM_SUBMITTER_NAME = "SubmitterName";
    public static final String PARAM_CLAIM_NAME = "ClaimId";
    public static final String PARAM_MODEL_RECORD_NAME = "ModelRecord";
    public static final String PARAM_MODEL_NAME = "Model";
    public static final String PARAM_MODEL_OBJECT_NAME = "ModelObject";

    public static final String HEADER_PHASE_PREPROCESSING = "PREPROCESSING";
    public static final String HEADER_PHASE_POSTPROCESSING = "POSTPROCESSING";
    public static final String HEADER_PHASE_PREPROCESSING_LABEL = "Preprocessing phase";
    public static final String HEADER_PHASE_POSTPROCESSING_LABEL = "Postprocessing phase";

    public static final String PARAM_COMPENSATION_TYPE_NAME = "CompensationTypeName";
    public static final String PARAM_COMPENSATION_RECORD_NAME = "CompensationRecordName";
    public static final String PARAM_COMPENSATION_NAME = "CompensationName";

    private static final ObjectMapper objectMapper = Utils.getObjectMapper();

    protected AbstractFormula onDiskLogic;
    protected Path logicDir;

    public AbstractLogicSerializer(AbstractFormula onDiskLogic) {
        super(false, true);
        this.onDiskLogic = onDiskLogic;
        if (onDiskLogic == null) {
            throw new IllegalArgumentException("LogicSerializer cannot be created with onDiskLogic=null");
        }
        logicDir = onDiskLogic.getDirPath();
    }


    public static String getFileName(String elementName, ElementCombinationType combinationType) {
        return elementName + "." + combinationType.getFileExtension();
    }

    public static String getFileNameForElement(AbstractFormulaElement element) {
        return getFileName(element.getElementName(), ElementCombinationType.valueOf(element.getCombinationType()));
    }

    public static String getTestFileNameForElement(AbstractFormulaElement element) {
        return getFileName(element.getElementName() + "Test", ElementCombinationType.valueOf(element.getCombinationType()));
    }

    public static String getFileExtensionForElement(AbstractFormulaElement element) {
        return ElementCombinationType.valueOf(element.getCombinationType()).getFileExtension();
    }

    public static ObjectMapper getObjectMapper() {
        return objectMapper;
    }

    public static Map<String, Object> convertParamListToMap(List<? extends ContextParameter> paramsList) {
        Map<String, Object> paramsMap = new HashMap<>();
        for (ContextParameter param : paramsList) {
            paramsMap.put(param.getName(), param.getValue());
        }
        return paramsMap;
    }

    public static BaseMap fetch(PfxClient client, TypeCode typeCode, Object... values) throws PfxClientException {
        switch (typeCode) {
            case MANUALPRICELISTITEM:
                return ManualPriceListApi.fetchManualPriceListItem(client, (long) values[0], PfxFilters.getSimpleCriteria(FIELD_SKU, values[1]));
            case PRICEGRIDITEM:
                return PriceGridApi.fetchPriceGridItem(client, (long) values[0], PfxFilters.getSimpleCriteria(FIELD_SKU, values[1]));
            case PRICELISTITEM:
                return PriceListApi.fetchPriceListItem(client, (long) values[0], PfxFilters.getSimpleCriteria(FIELD_SKU, values[1]));
            case MATRIXPRICEGRIDITEM:
                return PriceGridApi.fetchPriceGridItem(client, (long) values[0], PfxFilters.getSimpleCriteria2(FIELD_SKU, values[1], FIELD_CI_KEY2, values[2]));
            case MATRIXPRICELISTITEM:
                return PriceListApi.fetchPriceListItem(client, (long) values[0], PfxFilters.getSimpleCriteria2(FIELD_SKU, values[1], FIELD_CI_KEY2, values[2]));
            default:
                String[] identifierFieldNames = typeCode.getIdentifierFieldNames();
                if (identifierFieldNames.length > values.length) {
                    throw new IllegalArgumentException("TypeCode.getIdentifierFieldNames() for " + typeCode.name() + " is not yet supported");
                }

                return GeneralApi.fetchOne(client, typeCode.getShortName(), PfxFilters.getSimpleCriteriaN(identifierFieldNames, values));
        }
    }

    private static String getCachedTypedId(Map<String, Object> params) {
        String typedId = (String) params.get(PARAM_FM_ITEMS);
        if (typedId != null) {
            return typedId;
        }
        typedId = (String) params.get(PARAM_FM_ITEM);
        if (typedId != null) {
            return typedId;
        }
        typedId = (String) params.get(PARAM_WFBM_APPROVABLE);
        if (typedId != null) {
            return typedId;
        }
        return null;
    }

    public static class CachedEntry {
        PfxClient client = null;
        Map<TypeCode, Object[]> values = new HashMap<>();
        Map<TypeCode, String> typedIds = new HashMap<>();

        public boolean equals(PfxClient client, TypeCode typeCode, Object[] values) {
            return ((this.client == null && client == null) || (this.client != null && this.client.equals(client)))
                    && Arrays.equals(values, this.values.get(typeCode));
        }

        public String set(PfxClient client, TypeCode typeCode, Object[] values, String typedId) {
            this.client = client;
            this.values.put(typeCode, values);
            this.typedIds.put(typeCode, typedId);
            return typedId;
        }

        public String get(PfxClient client, TypeCode typeCode, Object[] values) {
            if (equals(client, typeCode, values)) {
                return typedIds.get(typeCode);
            }
            return null;
        }
    }

    private static String fetchId(PfxClient client, TypeCode typeCode, Object... values) throws PfxClientException {
        BaseMap item = fetch(client, typeCode, values);

        if (item == null) {
            throw new PfxClientException(typeCode.name() + " object was not found");
        }

        return (String) item.get("typedId");
    }

    private static String fetchCached(AbstractLogicSerializer.CachedEntry cachedEntry, PfxClient client, TypeCode typeCode, Object... values) throws PfxClientException {
        if (cachedEntry != null) {
            String typedId = cachedEntry.get(client, typeCode, values);
            if (typedId != null) {
                return typedId;
            }
        }
        return cachedEntry.set(client, typeCode, values, fetchId(client, typeCode, values));
    }

    private static String convertParameterStrings(AbstractLogicSerializer.CachedEntry cachedEntry, PfxClient client, TypeCode typeCode, Map<String, Object> params, String... paramNames) throws PfxClientException {
        Object[] values = new Object[paramNames.length];

        if (values.length > 0 && values[0] instanceof String && StringUtils.isEmpty((String) values[0])) {
            return null;
        }

        for (int i = 0; i < paramNames.length; i++) {
            values[i] = params.remove(paramNames[i]);
        }

        return fetchCached(cachedEntry, client, typeCode, values);
    }

    private static String convertParameterIdStrings(AbstractLogicSerializer.CachedEntry cachedEntry, PfxClient client, TypeCode typeCode, Map<String, Object> params, String paramName, Object... values) throws PfxClientException {
        Long value = (Long) params.get(paramName);

        if (value == null || value == 0) {
            return null;
        }

        return fetchCached(cachedEntry, client, typeCode, ArrayUtils.addAll(new Object[]{value}, values));
    }

    private static void convertRebateDates(Map<String, Object> params) {
        Map<String, Object> nature = new LinkedHashMap<>();
        Object param = params.remove(FIELD_RBA_STARTDATE);
        nature.put(FIELD_RBA_STARTDATE, param);
        param = params.remove(FIELD_RBA_ENDDATE);
        nature.put(FIELD_RBA_ENDDATE, param);
        param = params.remove(FIELD_RBA_PAYOUTDATE);
        nature.put(FIELD_RBA_PAYOUTDATE, param);
        param = params.remove(FIELD_RBA_CALCULATIONDATE);
        nature.put(FIELD_RBA_CALCULATIONDATE, param);
        params.put(PARAM_FM_NATURE, nature);
    }

    private static void convertCompensationDates(Map<String, Object> params) {
        Map<String, Object> nature = new LinkedHashMap<>();
        Object param = params.remove(FIELD_SCO_STARTDATE);
        nature.put(FIELD_SCO_STARTDATE, param);
        param = params.remove(FIELD_SCO_ENDDATE);
        nature.put(FIELD_SCO_ENDDATE, param);
        param = params.remove(FIELD_SCO_PAYOUTDATE);
        nature.put(FIELD_SCO_PAYOUTDATE, param);
        param = params.remove(FIELD_SCO_CALCULATIONDATE);
        nature.put(FIELD_SCO_CALCULATIONDATE, param);
        params.put(PARAM_FM_NATURE, nature);
    }

    private static String convertParametersObjectExtension(PfxClient client, Map<String, Object> params, String keyField, Object keyFieldValue, TypeCode typeCode) throws PfxClientException {
        String pxName = (String) params.remove(PARAM_STUDIO_OX_NAME);

        if (!StringUtils.isEmpty(pxName)) {
            And condition = new And(
                    equal(FIELD_NAME, pxName),
                    equal(keyField, keyFieldValue)
            );

            for (String paramName : new HashSet<>(params.keySet())) {
                if (paramName != null && paramName.startsWith(PARAM_STUDIO_KEY)) {
                    int i = paramName.indexOf('_', PARAM_STUDIO_KEY.length());
                    String fieldName = paramName.substring(i + 1);

                    Object value = params.remove(paramName);
                    condition.addCriterion(equal(fieldName, value));
                }
            }

            String typedId = getCachedTypedId(params);
            if (typedId != null) {
                return typedId;
            }

            BaseMap item = GeneralApi.fetchOne(client, typeCode.getShortName(), condition);
            if (item == null) {
                String msg = String.format("%s object was not found", typeCode.name());
                throw new PfxClientException(msg);
            }

            return (String) item.get("typedId");
        }

        return null;
    }

    private static String convertParametersPaDataload(PfxClient client, Map<String, Object> params, MetadataCache metadataProvider, TypeCode typeCode) throws PfxClientException {
        String datamartName = (String) params.remove(PARAM_STUDIO_DS_NAME);

        if (datamartName != null) {
            Map<String, CriteriaList> filterMap = new HashMap();

            for (String paramName : new HashSet<>(params.keySet())) {
                if (paramName.startsWith(PARAM_STUDIO_KEY)) {
                    int i = paramName.indexOf('_', PARAM_STUDIO_KEY.length());
                    String indexStr = paramName.substring(PARAM_STUDIO_KEY.length(), i);
                    String name = paramName.substring(i + 1);
                    Object value = params.remove(paramName);
                    if (!filterMap.containsKey(indexStr)) {
                        filterMap.put(indexStr, new And());
                    }
                    filterMap.get(indexStr).addCriterion(equal(name, value));
                }
            }

            Map<String, Object> datamart = metadataProvider.getMetasDatamartByName(typeCode, datamartName);
            if (datamart != null) {
                String typedId = (String) datamart.get(FIELD_TYPEDID);
                List<BaseMap> records = DatamartApi.fetchDatamart(client, typedId, 0, filterMap.size(), null, null, new Or(filterMap.values()));
                HashMap<String, Object> nature = new HashMap<>();
                nature.put(PARAM_FM_SELECTEDROWS, records);
                params.put(PARAM_FM_NATURE, nature);
                return typedId;
            }
        }

        return null;
    }

    public static void convertParameters(AbstractLogicSerializer.CachedEntry cache, PfxClient client, Map<String, Object> params, AbstractFormula logic, MetadataCache metadataProvider) throws PfxClientException {
        boolean isWorkflowLogic = logic.isWorkflowLogic();

        String typedId = null;

        Object skusObject = params.remove(CONTEXTPARAM_NAME_PRODUCTSKU);
        List<String> skus;
        if (skusObject instanceof List) {
            skus = (List) skusObject;
        } else {
            skus = new ArrayList<>();
            skus.add((String) skusObject);
        }
        Object typeCodeObj = params.remove(PARAM_STUDIO_TYPE_CODE);
        TypeCode typeCode = typeCodeObj instanceof TypeCode ? (TypeCode) typeCodeObj : typeCodeObj instanceof String ? TypeCode.valueOf((String) typeCodeObj) : null;

        if (isWorkflowLogic) {
            String submitterTypedId = convertParameterStrings(cache, client, USER, params, PARAM_SUBMITTER_NAME);
            params.put(PARAM_WFBM_SUBMITTER, submitterTypedId);
        }
        List typedIds = new ArrayList();
        for (String sku : skus) {
            if (typeCode == QUOTE) {
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_QUOTE_NAME);
            } else if (typeCode == CONTRACT) {
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_CONTRACT_NAME);
            } else if (typeCode == REBATEAGREEMENT) {
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_REBATE_AGREEMENT_NAME);
            } else if (typeCode == PRODUCT && sku != null && !sku.isEmpty()) {
                typedId = fetchCached(cache, client, typeCode, sku);
            } else if (typeCode == PRICELIST) {
                typedId = convertParameterIdStrings(cache, client, typeCode, params, PARAM_PRICE_LIST_ID);
            } else if (typeCode == REBATETYPE) {
                convertRebateDates(params);
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_REBATE_TYPE_NAME);
            } else if (typeCode == REBATERECORD) {
                convertRebateDates(params);
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_REBATE_RECORD_NAME);
            } else if (typeCode == CONTRACTTERMTYPE) {
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_CONTRACT_TERM_TYPE_NAME);
            } else if (typeCode == DMDATAMART || typeCode == DMDATASOURCE || typeCode == DMDATAFEED) {
                typedId = convertParametersPaDataload(client, params, metadataProvider, typeCode);
            } else if (typeCode == PRICERECORD) {
                typedId = convertParameterStrings(cache, client, typeCode, params, FIELD_PR_SOURCEID, FIELD_PR_LINEID);
            } else if (typeCode == DATACHANGEREQUEST) {
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_DATA_CHANGE_REQUEST_NAME);
            } else if (typeCode == CLAIM) {
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_CLAIM_NAME);
            } else if (typeCode == MODELRECORD) {
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_MODEL_RECORD_NAME);
            } else if (typeCode == PRICEOPTIMIZATIONMODEL) {
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_MODEL_NAME);
            } else if (typeCode == MODELOBJECT) {
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_MODEL_OBJECT_NAME);
            } else if (typeCode == COMPENSATION) {
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_COMPENSATION_NAME);
            } else if (typeCode == COMPENSATIONCONDITIONTYPE) {
                convertCompensationDates(params);
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_COMPENSATION_TYPE_NAME);
            } else if (typeCode == COMPENSATIONRECORD) {
                convertCompensationDates(params);
                typedId = convertParameterStrings(cache, client, typeCode, params, PARAM_COMPENSATION_RECORD_NAME);
            } else if (!StringUtils.isEmpty(sku)) {
                String key2 = (String) params.remove(PARAM_SECONDARY_KEY);

                if (typeCode == PRICEGRIDITEM) {
                    typedId = convertParameterIdStrings(cache, client, typeCode, params, PARAM_PRICE_GRID_ID, sku);
                } else if (typeCode == MATRIXPRICEGRIDITEM && key2 != null) {
                    typedId = convertParameterIdStrings(cache, client, typeCode, params, PARAM_PRICE_GRID_ID, sku, key2);
                } else if (typeCode == PRICELISTITEM) {
                    typedId = convertParameterIdStrings(cache, client, typeCode, params, PARAM_PRICE_LIST_ID, sku);
                } else if (typeCode == MATRIXPRICELISTITEM && key2 != null) {
                    typedId = convertParameterIdStrings(cache, client, typeCode, params, PARAM_PRICE_LIST_ID, sku, key2);
                } else if (typeCode == MANUALPRICELISTITEM) {
                    typedId = convertParameterIdStrings(cache, client, typeCode, params, PARAM_MANUAL_PRICE_LIST_ID, sku);
                } else if (typeCode == PRODUCTEXTENSION) {
                    typedId = convertParametersObjectExtension(client, params, FIELD_SKU, sku, typeCode);
                }
            } else if (typeCode == CUSTOMER || typeCode == CUSTOMEREXTENSION) {
                if (typeCode == CUSTOMER) {
                    typedId = convertParameterStrings(cache, client, typeCode, params, CONTEXTPARAM_NAME_CUSTOMER);
                } else {
                    String customerId = (String) params.remove(CONTEXTPARAM_NAME_CUSTOMER);
                    typedId = convertParametersObjectExtension(client, params, FIELD_CUSTOMER_ID, customerId, typeCode);
                }
            }
            typedIds.add(typedId);
        }

        if (typedIds.size() > 1) {
            params.put(PARAM_FM_ITEMS, typedIds);
        } else if (typedId != null) {
            params.put(isWorkflowLogic ? PARAM_WFBM_APPROVABLE : PARAM_FM_ITEM, typedId);
        }

        removeParamsAfterConversion(params);
    }

    public static void removeParamsAfterConversion(Map<String, Object> params) {
        params.remove(PARAM_PRICE_LIST_ID);
        params.remove(PARAM_PRICE_GRID_ID);
        params.remove(PARAM_MANUAL_PRICE_LIST_ID);
    }

    public void createElement(AbstractFormulaElement element, int index, boolean overwrite) throws IOException {
        List<AbstractFormulaElement> elements = onDiskLogic.getElements();

        int existingElement = onDiskLogic.getElementIndexByName(element.getElementName());
        boolean elementAlreadyExists = existingElement != -1;

        if (!overwrite && elementAlreadyExists) {
            String msg = String.format("Element with name '%s' exists already in logic '%s'!", element.getElementName(), onDiskLogic.getUniqueName());
            throw new IOException(msg);
        }

        writeElementAsFile(element);

        // formulaExpression should not be in the logic.json, nor in ObjectCache
        element.setFormulaExpression(null);

        if (overwrite && elementAlreadyExists) {
            elements.remove(existingElement);
        }
        elements.add(index, element);

        writeLogicJsonFile();
    }

    public static ReplyingSupplierFactory getReplyingSupplierFactory(TypeDescriptor typeDescriptor, Path modulePath, ProcessingContext context, ObjectNode node, ItemFilter filter) throws ProcessingException {
        try {
            SupplierFactory supplierFactory = new FS_SupplierFactory(modulePath).with(new FileSystemOpsBase(false, true));
            ReplyingSupplierFactory replyingSupplierFactory = new ReplyingSupplierFactory();

            Transformations transformations = new Transformations(supplierFactory, replyingSupplierFactory.consumerFactory());
            if (node != null) {
                transformations.withSingleItemFilter(typeDescriptor.getName(), node);
            }
            transformations.withFilter(filter)
                    .with(context)
                    .transform();
            return replyingSupplierFactory;
        } catch (IOException | URISyntaxException e) {
            BusinessKey businessKey = typeDescriptor.businessKey(node);
            String businessKeyString = businessKey.toString();
            throw new ProcessingException(node, String.format("Read of %s %s failed", typeDescriptor.getName(), businessKeyString), e);
        }
    }

    public static AbstractFormula readLogic(Path modulePath, ProcessingContext context, Path relPath, boolean isShallow, boolean adjustLabelTranslations) throws ProcessingException {
        TypeDescriptor typeDescriptor = TypeDescriptors.fromPath(relPath);
        List<ObjectNode> objects = readObjectsonPath(context, modulePath, relPath, isShallow);
        if (adjustLabelTranslations) {
            JsonNode elements = objects.get(0).path(FIELD_FE_ELEMENTS);
            if (elements.isArray()) {
                ArrayNode array = (ArrayNode) elements;
                for (int i = 0; i < array.size(); i++) {
                    array.set(i, new UnfoldField(context, FIELD_LABELTRANSLATIONS).apply((ObjectNode) array.get(i)));
                }
            }

        }
        return AbstractFormula.createLogic(objects.get(0), typeDescriptor, modulePath);
    }

    public static List<ObjectNode> readObjectsonPath(ProcessingContext context, Path dir, Path relPath, boolean isShallow) throws ProcessingException {
        TypeDescriptor typeDescriptor = TypeDescriptors.fromPath(relPath);
        try {
            SupplierFactory supplierFactory = new FS_SupplierFactory(dir).with(new FileSystemOpsBase(false, true));

            ReplyingSupplierFactory replyingSupplierFactory = new ReplyingSupplierFactory();

            new Transformations(supplierFactory, replyingSupplierFactory.consumerFactory())
                    .with(context)
                    .shallow(isShallow)
                    .withPathFilter(relPath)
                    .transform();

            List<ObjectNode> objects = replyingSupplierFactory.list(typeDescriptor.getName());
            if (objects.size() < 1) {
                throw new ProcessingException(objects, String.format("Cannot read %s from %s", typeDescriptor.getName(), dir.toString()), null);
            }
            return objects;
        } catch (IOException | URISyntaxException e) {
            throw new ProcessingException(null, String.format("Cannot read %s from %s", typeDescriptor.getName(), dir.toString()), e);
        }
    }

    public String toString() {
        return logicDir == null ? super.toString() : logicDir.toString();
    }

    public abstract void writeLogicJsonFile() throws IOException;

    public abstract void writeElementAsFile(AbstractFormulaElement element) throws IOException;

    public abstract String[] readParamFiles();

    public abstract <T extends ContextParameter> List<T> readParamsFile(String fileName) throws IOException;

    public abstract void deleteParamsFile(String fileName) throws IOException;

    public static List<String> getAllMatchedLibraries(Path calculationLogicDir, String libraryName) throws IOException {
        List<String> matchedLibraries = new ArrayList<>();
        Stream<Path> paths = Files.find(calculationLogicDir,
                1,
                (path, basicFileAttributes) ->
                        path.toFile().isDirectory() && (
                                path.getFileName().toString().matches(
                                        libraryName + "\\+([12]\\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[12]\\d|3[01]))") ||
                                        libraryName.equalsIgnoreCase(path.getFileName().toString()))

        );

        paths.forEach(path -> matchedLibraries.add(path.getFileName().toString()));

        return matchedLibraries;
    }

    public static AbstractFormula getLatestLibrary(Path modulePath, String uniqueName, List<String> libraryNames) {
        Path calculationLogicDir = modulePath.resolve(TransformCalculationLogic.DIRNAME);

        //always get the latest library if they are of the same name, different valid after date
        Collections.sort(libraryNames);

        AbstractFormula latestFormula = null;

        for (String libraryName : libraryNames) {
            Path libraryPath = calculationLogicDir.resolve(libraryName);
            Path relPath = modulePath.relativize(libraryPath);

            try {
                AbstractFormula formula = LogicSerializer.readLogic(modulePath,
                        new ProcessingContextImpl(), relPath, true, false);

                if (formula.getFormulaNatureObject() == FormulaNature.LIBRARY &&
                        formula.getUniqueName().equalsIgnoreCase(uniqueName)) {
                    LocalDate validAfter = LocalDate.parse(formula.getValidAfter());

                    if (validAfter.isBefore(LocalDate.now()) || validAfter.isEqual(LocalDate.now())) {
                        latestFormula = formula;
                    }

                }

            } catch (ProcessingException e) {
                log.warn("Ignored Library:" + libraryName, e);
            }

        }

        return latestFormula;

    }

    public static boolean isLogicFolder(String name) {
        return TransformCalculationLogic.DIRNAME.equals(name) || TransformWorkflowFormula.DIRNAME.equals(name);
    }
}
